/*
 * libbsr - allocation services for the Barrier Synchronization Register
 *          on IBM POWER systems
 *
 * Copyright (C) 2008 Sonny Rao, IBM Corporation.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "bsr.h"

void __bsr_fini() __attribute__((destructor));

static char *bsr_lockdir  = "/var/lib/bsr/";
static char *bsr_sysfsdir = "/sys/class/bsr/";

struct bsr_lockinfo {
	pid_t		pid;
};

struct bsr_info {
	unsigned num_devs;
	unsigned max_size;
	unsigned num_sizes;
	unsigned sizemap;
	unsigned free;
	unsigned freemap;
/* enumerate all power of 2 sizes up to 8 MB
 * so we can keep a count of how many there are
 */
#define MAX_BSR_SIZE 23
/* support up to 256 (8 bits) of each type*/
#define MAX_CLASS_DEVS 256
	unsigned counts[MAX_BSR_SIZE + 1];
};

/* we'll cache this info */
static struct bsr_info bsr_all;

/* This tells us that there are two BSR sizes, one is 8 bytes and the other
 * is larger and the 8 byte BSRs map ontop of the larger one so if the big
 * one is taken, the smaller ones are also taken. This is the source of much
 * ugliness in the code.
 */
static unsigned bsr_alias_mode;
static unsigned bsr_big_size;

/* our idea of "total bytes" */
static unsigned bsr_total_bytes;

#define mk_bsr_id(_class, _id)		(((_class) << 24) | (_id))
#define bsr_id_to_class(_bsr_id)	((_bsr_id) >> 24)
#define bsr_id_to_id(_bsr_id)		((_bsr_id) & 0xff)
#define bsr_id_to_bytes(_bsr_id)	(1 << bsr_id_to_class(_bsr_id))

static int bsr_log2(unsigned num)
{
	int ret = 0;

	if (!num)
		return ret;

	while (num) {
		ret++;
		num = num >> 1;
	}

	return ret - 1;
}

static unsigned bsr_lockname_to_bsr_id(char *bsr_lockname)
{
	unsigned bytes, id, class;
	int count = sscanf(bsr_lockname, "bsr%d_%d", &bytes, &id);
	if (count != 2)
		return -1;

	class = bsr_log2(bytes);
	return mk_bsr_id(class, id);
}

static int read_bsr_lock(unsigned bsr_id, struct bsr_lockinfo *li)
{
	char buf[64];
	FILE *f;
	int rc;

	sprintf(buf, "%s/bsr%d_%d", bsr_lockdir, bsr_id_to_bytes(bsr_id),
		bsr_id_to_id(bsr_id));
	f = fopen(buf, "r");
	if (!f)
		return -1;

	rc = fread(li, sizeof(*li), 1, f);
	fclose(f);

	return rc ? 0 : -1;
}

static int read_bsr_lock_by_name(char *bsr_lockname, struct bsr_lockinfo *li)
{
	unsigned bsr_id = bsr_lockname_to_bsr_id(bsr_lockname);
	return read_bsr_lock(bsr_id, li);
}

static int write_bsr_lock(unsigned bsr_id, struct bsr_lockinfo *li)
{
	char buf[32];
	int fd, rc;
	unsigned id = bsr_id_to_id(bsr_id);
	unsigned bytes = bsr_id_to_bytes(bsr_id);

	snprintf(buf, 32, "%sbsr%d_%d", bsr_lockdir, bytes, id);
	fd = open(buf, O_WRONLY, S_IRWXU);
	if (fd < 0)
		return -1;

	rc = write(fd, li, sizeof(*li));
	close(fd);

	return (rc > 0) ? 0 : -1;
}

static int delete_bsr_lock(unsigned bsr_id)
{
	char buf[64];
	unsigned bytes = bsr_id_to_bytes(bsr_id);
	unsigned id = bsr_id_to_id(bsr_id);

	snprintf(buf, 64, "%sbsr%d_%d", bsr_lockdir, bytes, id);
	return unlink(buf);
}

static int create_bsr_lock(unsigned bsr_id)
{
	char buf[32];
	int fd, rc;
	struct bsr_lockinfo li;
	unsigned bytes = bsr_id_to_bytes(bsr_id);
	unsigned id = bsr_id_to_id(bsr_id);

	snprintf(buf, 32, "%sbsr%d_%d", bsr_lockdir, bytes, id);
	fd = open(buf, O_CREAT | O_EXCL | O_RDWR,  S_IRWXU);
	if (fd < 0)
		return -1;

	li.pid = getpid();
	rc = write(fd, &li, sizeof(li));
	close(fd);

	if (rc < 1) {
		delete_bsr_lock(bsr_id);
		return rc;
	}

	return 0;
}

static int bsr_query_dir(char *dirname, struct bsr_info *bsr_info)
{
	struct dirent *ent;
	DIR *sysdir;

	sysdir = opendir(dirname);
	if (!sysdir)
		return ENODEV;

	memset(bsr_info, 0, sizeof(*bsr_info));

	while ((ent = readdir(sysdir)) != NULL) {
		unsigned size, id, class;
		int count = sscanf(ent->d_name, "bsr%d_%d", &size, &id);
		if (count == 2) {
			bsr_info->num_devs++;
			if (bsr_info->max_size < size)
				bsr_info->max_size = size;

			class = bsr_log2(size);
			bsr_info->counts[class]++;

			if (!(bsr_info->sizemap & (1 << class))) {
				bsr_info->num_sizes++;
				bsr_info->sizemap |= (1 << class);
			}
		}
	}

	closedir(sysdir);
	return 0;
}

static unsigned total_counts(struct bsr_info *bsi)
{
	int i;
	unsigned total = 0;

	for (i = 0; i < MAX_BSR_SIZE; i++) {
		unsigned bytes = 1 << i;
		unsigned all = bsi->counts[i] * bytes;
		total += all;
	}

	return total;
}

int bsr_query(unsigned *total_bytes, uint32_t *supp_window_mask,
	      unsigned *free_bytes, uint32_t *free_window_mask)
{
	int ret;
	struct bsr_info bsr_locks;

	/* only read the total info once */
	if (!bsr_all.num_devs) {
		ret = bsr_query_dir(bsr_sysfsdir, &bsr_all);
		if (ret)
			return ret;

		/* if there are two sizes and one is 8bytes
		 * the smaller ones map over the big one
		 */
		if ((bsr_all.num_sizes == 2) && (bsr_all.sizemap & (1 << 3))) {
			bsr_alias_mode = 1;
			bsr_big_size  = bsr_all.sizemap & (~(1 << 3));
			bsr_total_bytes = bsr_big_size;
		} else {
			bsr_total_bytes = total_counts(&bsr_all);
		}
	}

	*supp_window_mask = bsr_all.sizemap;
	*total_bytes = bsr_total_bytes;

	/* look at the lockdir and figure out how many are free */
	ret = bsr_query_dir(bsr_lockdir, &bsr_locks);
	if (ret)
		return ret;

	/* alias mode is messy.. if the big one is taken, they're
	 * all taken, also if any small one is taken, then the big one is
	 * also taken.
	 */
	if (bsr_alias_mode) {
		if (bsr_locks.sizemap & bsr_big_size) {
			*free_bytes = 0;
			*free_window_mask = 0;
		} else if (bsr_locks.sizemap & 8) {
			/* someone has a small one */
			*free_bytes = bsr_big_size - total_counts(&bsr_locks);
			*free_window_mask = 8;
		} else {
			/* everything is free */
			*free_bytes = bsr_total_bytes;
			*free_window_mask = bsr_all.sizemap;
		}
	} else {
		int i;
		unsigned free_mask  = 0;

		*free_bytes = bsr_total_bytes - total_counts(&bsr_locks);
		for (i = 0; i < MAX_BSR_SIZE; i++) {
			if (bsr_all.counts[i]) {
				if (bsr_locks.counts[i] < bsr_all.counts[i])
					free_mask = 1 << i;
			}
		}

		*free_window_mask = free_mask;
	}

	return 0;
}

static int bsr_alloc_size(unsigned bytes)
{
	struct dirent *ent;
	DIR *dir;
	unsigned class, class_total, i;
	char tmpl[32];
	char alloced[MAX_CLASS_DEVS];

	class = bsr_log2(bytes);
	class_total = bsr_all.counts[class];

	memset(alloced, 0, sizeof(alloced));

	snprintf(tmpl, 32, "bsr%d_%%d", bytes);

	dir = opendir(bsr_lockdir);
	while ((ent = readdir(dir)) != NULL) {
		unsigned num;
		if (ent->d_type == DT_REG) {
			if (sscanf(ent->d_name, tmpl, &num) == 1)
				alloced[num] = 0xff;
		}
	}
	closedir(dir);

	for (i = 0; i < class_total; i++) {
		unsigned bsr_id;
		int rc;

		if (alloced[i])
			continue;

		bsr_id = mk_bsr_id(class, i);
		rc = create_bsr_lock(bsr_id);
		if (rc)
			return -EIO;
		else
			return bsr_id;
	}

	return -EBUSY;
}

static int bsr_get_attr(unsigned size, unsigned id, char *name,
			unsigned *bsr_attr)
{
	char buf[64];
	int fd;

	snprintf(buf,64, "%sbsr%d_%d/%s", bsr_sysfsdir, size, id, name);
	if ((fd = open(buf, O_RDONLY)) < 0)
		return errno;

	if (read(fd, buf, 64) < 0){
		close(fd);
		return EIO;
	}

	close(fd);
	sscanf(buf, "%d", bsr_attr);
	return 0;
}

int bsr_alloc(unsigned bsr_bytes, unsigned *bsr_stride, unsigned *bsr_id)
{
	unsigned total, allmask, free, freemask, my_bsr_id;
	int ret;

	ret = bsr_query(&total, &allmask, &free, &freemask);
	if (ret)
		return ret;

	if (!free || !(freemask & bsr_bytes))
		return EBUSY;

	my_bsr_id = bsr_alloc_size(bsr_bytes);
	if (my_bsr_id < 0)
		return my_bsr_id;

	ret = bsr_get_attr(bsr_bytes, bsr_id_to_id(my_bsr_id),
			   "bsr_stride", bsr_stride);
	if (ret) {
		bsr_free(my_bsr_id);
		return ret;
	}

	*bsr_id = my_bsr_id;
	return 0;
}

/* TODO, more checking ? */
int bsr_free(unsigned bsr_id)
{
	char buf[64];
	unsigned id = bsr_id_to_id(bsr_id);
	unsigned bytes = bsr_id_to_bytes(bsr_id);
	int ret;

	snprintf(buf, 64, "%sbsr%d_%d", bsr_lockdir, bytes, id);
	ret = unlink(buf);

	return ret;
}

void *bsr_map(void *addr, unsigned bsr_id, int prot, int flags,
	      unsigned *bsr_length)
{
	char buf[64];
	unsigned bytes = bsr_id_to_bytes(bsr_id);
	unsigned id = bsr_id_to_id(bsr_id);
	unsigned length = 0;
	int fd, ret;
	void *ret_addr;

	ret = bsr_get_attr(bytes, id, "bsr_length", &length);
	if (ret)
		return NULL;

	*bsr_length = length;

	snprintf(buf, 64, "%sbsr%d_%d", "/dev/", bytes, id);
	fd = open(buf, O_RDWR);
	if (fd < 0)
		return NULL;

	if (!prot)
		prot = PROT_READ | PROT_WRITE;
	if (!flags)
		flags = MAP_SHARED;

	ret_addr = mmap(addr, length, prot, flags, fd, 0);
	close(fd);

	return ret_addr;
}

int bsr_unmap(void *addr, unsigned length)
{
	return munmap(addr, length);
}

void __bsr_fini(void)
{
	pid_t pid;
	struct dirent *ent;
	DIR *dir;

	pid = getpid();

	dir = opendir(bsr_lockdir);
	if (!dir)
		return;

	while ((ent = readdir(dir)) != NULL) {
		struct bsr_lockinfo li;
		unsigned bsr_id;
		int rc = 0;

		if (ent->d_type != DT_REG)
			continue;

		rc = read_bsr_lock_by_name(ent->d_name, &li);
		if (rc)
			continue;

		if (li.pid == pid) {
			bsr_id = bsr_lockname_to_bsr_id(ent->d_name);
			if (bsr_id != -1)
				bsr_free(bsr_id);
		}
	}

	closedir(dir);
}
